
    #	    Create GIF animation of transit from extract CSV file
    
    #	by John Walker
    #	http://www.fourmilab.ch/
    
    #	For details, please see: http://www.fourmilab.ch/documents/canon_transits/

    use strict;
    use warnings;
    
    use Math::VectorReal;
    
    
    #	Set DemoJob to choose a standard demo job
    #	    0:  No demo job
    #	    1:  Transit of Mercury and Venus from Earth 69163-07-26
    #	    2:	Transit of Venus, Earth, and Moon from Saturn 23364-12-16
    
    my $DemoJob = 1;
    
    my $ifile = ''; 	    	# Input file name ($ARGV[0] if null)
    
    #	Set to limit portion of extract file rendered as animation
    my $jd_min = -1e100;
    my $jd_max = 1e100;
    
    #	View from this planet index (in @planets below)
    my $viewFrom = 3;	    	# Default is view from Earth
    
    #	Inflate the sizes of planets by this factor to render them more visible
    my $inflate_planets = 1;
#    my $inflate_planets = 4;
    
    #	Zoom the image by this factor
    my $zoom_sun = 2;
#    zoom_sun = 4;

    #	Make this type of frame file
    my $frametype = 'gif';  	    # "gif" or "png"
    
    #	Specify animation frame width and height
    my ($fwidth, $fheight) = (1024, 768);
    
    #	Time per frame in 1/100 seconds
    my $frameTime = 50;

    die("\$DemoJob = $DemoJob.  No extract file can be specified on command line")
    	if (($DemoJob > 0) && ($#ARGV >= 0));

    if ($DemoJob == 1) {
    	$ifile = '69163_extract.csv';
	$viewFrom = 3;
    	$jd_min = 26982532.85;	# 69163-07-26 transit of Venus and Mercury
    	$jd_max = 26982533.60;
    } elsif ($DemoJob == 2) {
    	$ifile = '23364_extract.csv';
	$viewFrom = 7;
    	$jd_min = 10254935.65;	# 23364-12-16 triple transit from Saturn
    	$jd_max = 10254936.44;
	($fwidth, $fheight) = (1200, 900);  # Make Moon big enough to see
    } elsif ($DemoJob != 0) {
    	die("Undefined \$DemoJob setting of $DemoJob");
    }

    my $povdir = "pov";     	    # POV scene definition directory

    my $framedir = "frames";   	    # Frame directory
    
    #	Gifsicle command to assemble animation
    my $gs = "gifsicle --colors 256 --delay $frameTime --loopcount=forever";

    #	POV-Ray rendering options
    my $pvopts = "-v -w$fwidth -h$fheight -d -q9";
    
    my $au = 149597870691.0;	    	# Astronomical unit in metres
        
    my @planets = (
    	"Sun",	    	    #  0
    	"Mercury",  	    #  1
	"Venus",    	    #  2
    	"Earth",    	    #  3
	"Moon",     	    #  4
	"Mars",     	    #  5
	"Jupiter",  	    #  6
	"Saturn",   	    #  7
	"Uranus",   	    #  8
     	"Neptune",  	    #  9
	"Pluto" );  	    # 10

    my %planet_col = (
    	"Sun", "Yellow",
    	"Mercury", "Green",
	"Venus", "White",
    	"Earth", "Blue",
	"Moon", "Red",
	"Mars", "Red",
	"Jupiter", "Green",
	"Saturn", "Yellow",
	"Uranus", "Cyan",
     	"Neptune", "Cyan",
	"Pluto", "Gray50" );
	
    my %planet_radius = (
    	"Sun", 6.96e8 / $au,
    	"Mercury", 2439700 / $au,
	"Venus", 6051800 / $au,
    	"Earth", 4.26352325064817808471E-5,
	"Moon", 1.16178124180819699985E-5,
	"Mars", 3397000 / $au,
	"Jupiter", 71492000 / $au,
	"Saturn", 60268000 / $au,
	"Uranus", 25559000 / $au,
     	"Neptune", 24764000 / $au,
	"Pluto", 1195000 / $au );
	
    my $l;
    
    my ($jd, $yy, $mo, $dd, $hh, $mm, $ss);
    my %planet_vec;
    my $ijd = 0;
    
    if ($ifile eq '') {
    	if ($#ARGV < 0) {
	    print(STDERR "Usage: perl tranimator.pl extract_file.csv\n");
	    exit(2);
	}
	$ifile = $ARGV[0];
    }
    
    open(FI, "<$ifile") || die("Cannot open input extract file $ifile");
    
round:
    while ($l = <FI>) {

	#   Load a state vector snapshot into memory
	
	$l =~ m/JD,([^,]+),([^,]+),([^,]+),([^,]+),([^,]+),([^,]+),([^,]+)\s/;
    	($jd, $yy, $mo, $dd, $hh, $mm, $ss) = ($1, $2, $3, $4, $5, $6, $7);
	
	while (1) {
	    $l = <FI>;
	    if ($l =~ m/^END/) {
	    	last;
	    }
	    $l =~ m/^(\w+),([\d\-\.e]+),([\d\-\.e]+),([\d\-\.e]+)/;
	    $planet_vec{$1} = vector($2, $3, $4);
	}
	if ($jd >= $jd_min) {
	    if ($jd > $jd_max) {
		last round;
	    }

	    if ($ijd == 0) {
		$ijd = int($jd);
	    }
    	    my $fname = sprintf("%.2f.pov", $jd);
	    genscene($fname, $viewFrom);
	    renderscene($fname);
	}
    }
    close(FI);
    	
    createanimation($ijd);

        
    #	Generate the POV-Ray scene definition file
    
    sub genscene {
    	my ($fname, $vfi) = @_;
	
	my $vfrom = $planets[$vfi];
	my $from_pos = outpos($planet_vec{$vfrom});
	my $sun_pos =  outpos($planet_vec{'Sun'});
	
	#   This is a rather tacky way to more or less align the horizontal axis of
	#   the image with the ecliptic.  This is nontrivial, since the origin of
	#   co-ordinates is the solar system barycentre, which is usually substantially
	#   displaced from the centre of the Sun.  We take the cross product of
	#   Sun->Earth and Sun->Jupiter vectors, then normalise and call that
	#   perpendicular to the ecliptic. Since the inclination of the orbit
	#   of Jupiter is only 1.3 degrees, this is close enough for the purposes
	#   of image synthesis. 
    	my $ecliptic_pole = (($planet_vec{'Earth'} - $planet_vec{'Sun'}) x ($planet_vec{'Jupiter'} - $planet_vec{'Sun'}))->norm();
	my $fvup = "< " . $ecliptic_pole->x() . ", " . $ecliptic_pole->y() . ", " . $ecliptic_pole->z() . " >";
	
	#   Compute the viewing angle so that the Sun is of uniform size regardless
	#   of where we're looking from.
	my $pi = 4 * atan2(1, 1);
	my $vangle = (8 * (180 / $pi) *
	    	    	atan2($planet_radius{'Sun'},
			($planet_vec{$vfrom} - $planet_vec{'Sun'})->length())) / $zoom_sun;
	
	open(OF, ">$povdir/$fname");
	
	print OF << "EOD";

#include "colors.inc"

    camera {
    	angle $vangle
	location $from_pos
	sky $fvup
	look_at $sun_pos
    }

    light_source {
    	$from_pos
	color White
    }
    
EOD

	for (my $i = $vfi - 1; $i >= 0; $i--) {
    	    my $pname = $planets[$i];
	    my $ppos = outpos($planet_vec{$pname});
	    my $psize = $planet_radius{$pname} * (($i > 0) ? $inflate_planets : 1);
	    print OF << "EOD";
    //  $pname
    sphere {
    	$ppos, $psize
	texture {
	    pigment {
	    	color $planet_col{$pname}
	    }
	}
    }
    
EOD
	}
	close(OF);
	
    }
    
    #	Render the scene for a given instant
    
    sub renderscene {
    	my ($fname) = @_;
	
	my $ofn = $fname;
	$ofn =~ s/pov$/$frametype/;
	my $flabel = sprintf("%.2f  %4d-%02d-%02d %02d:%02d", $jd, $yy, $mo, $dd, $hh, $mm);
	my $ly = $fheight - 16;
	my $c = "povray $povdir/$fname $pvopts +fp +o- | " .
		"ppmlabel -x 10 -y $ly -size 10 -text '$flabel' | " .
		(($frametype eq 'png') ?
    	    	    "pnmtopng" :	# For high-quality PNG still images
    	    	    "ppmquant 256 | ppmtogif") .
		">$framedir/$ofn\n";
	command($c);
	$gs .= " $framedir/$ofn";
    }
    
    #	Create an animation from the individual scene files
    
    sub createanimation {
    	my ($fjd) = @_;
	
	if ($frametype eq 'gif') {
	    my $c = "$gs >tranimator_$fjd.gif\n";
	    command($c);
	} else {
	    print(STDERR "Cannot create animation when \$frametype is \"$frametype\".\n");
	}
    }
    
    #	Edit a vector to POV-Ray format

    sub outpos {
    	my ($p) = @_;
	
	return "< " . $p->x() . ", " . $p->y() . ", " . $p->z(). " >";
    }

    #	Print a command and then run it
    
    sub command {
    	my ($c) = @_;
	
	print($c);
	system($c);
    }
